# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated.

"""Template for histogram stretch tools."""

from ..gtk.customwidgets import Label, HBox, VBox, Grid, Notebook
from ..gtk.keyboard import decode_key
from ..base import FigureCanvas, BaseToolbar, Container
from ..toolmanager import BaseToolWindow
from ..misc.utils import histogram_bins, plot_histograms, update_histograms, highlight_histogram, stats_string
import numpy as np
from matplotlib.figure import Figure
import matplotlib.ticker as ticker

class StretchTool(BaseToolWindow):
  """Histogram stretch tool class."""

  # Build window.

  _window_name_ = "" # Window name.

  def open(self, image):
    """Open tool window for image 'image'."""
    if not super().open(image, self._window_name_): return False
    wbox = VBox()
    self.window.add(wbox)
    fbox = VBox(spacing = 0)
    wbox.pack(fbox, expand = True, fill = True)
    self.widgets.fig = Figure(figsize = (10., 6.), layout = "constrained")
    canvas = FigureCanvas(self.widgets.fig)
    canvas.set_size_request(800, 480)
    fbox.pack(canvas, expand = True, fill = True)
    toolbar = BaseToolbar(canvas, self.widgets.fig)
    fbox.pack(toolbar)
    wbox.pack("Press [L] to toggle lin/log scale, [C] to plot the contrast enhancement function log(f')")
    grid = Grid()
    wbox.pack(grid, expand = True, fill = True)
    self.widgets.refstats = Label()
    grid.attach("<b>Reference</b>", 0, 0)
    grid.attach(self.widgets.refstats, 1, 0)
    self.widgets.imgstats = Label()
    grid.attach("<b>Image</b>", 0, 1)
    grid.attach(self.widgets.imgstats, 1, 1)
    options = self.options_widgets(self.widgets)
    if options is not None: wbox.pack(options)
    self.reference.stats = self.reference.statistics(channels = "RGBSVL")
    self.image.stats = self.reference.stats.copy()
    self.statchannels = ""     # Keys for the image statistics.
    self.histchannels = ""     # Keys for the image histograms.
    self.histcolors = []       # Colors of the image histograms.
    self.channelkeys = []      # Keys of the different channels/tabs.
    self.widgets.channels = {} # Widgets of the different channels/tabs.
    self.widgets.rgbtabs = Notebook()
    wbox.pack(self.widgets.rgbtabs)
    for key, name, color, lcolor in (("R", "Red", (1., 0., 0.), (1., 0., 0.)),
                                     ("G", "Green", (0., 1., 0.), (0., 1., 0.)),
                                     ("B", "Blue", (0., 0., 1.), (0., 0., 1.)),
                                     ("S", "HSV saturation", (1., .5, 0.), (1., .5, 0.)),
                                     ("V", "HSV value = max(RGB)", (0., 0., 0.), (1., 1., 1.)),
                                     ("L", "Luma", (.5, .5, .5), (1., 1., 1.))):
      channel = Container()
      tab = self.tab_widgets(key, channel)
      if tab is not None:
        channel.color = np.array(color)
        channel.lcolor = np.array(lcolor)
        self.statchannels += key
        self.histchannels += key
        self.histcolors.append(channel.color)
        self.channelkeys.append(key)
        self.widgets.channels[key] = channel
        self.widgets.rgbtabs.append_page(tab, Label(name))
    self.widgets.rgbtabs.connect("switch-page", lambda tabs, tab, itab: self.update("tab", tab = itab))
    wbox.pack(self.tool_control_buttons())
    if "R" not in self.histchannels: # Always include red, blue, green...
      self.histchannels += "R"
      self.histcolors.append(np.array((1., 0., 0.)))
    if "G" not in self.histchannels: # ...in the histograms even though...
      self.histchannels += "G"
      self.histcolors.append(np.array((0., 1., 0.)))
    if "B" not in self.histchannels: # ...there are no tabs for these channels.
      self.histchannels += "B"
      self.histcolors.append(np.array((0., 0., 1.)))
    self.histlogscale = False # Plot histograms with y log scale.
    self.histbins = histogram_bins(self.reference.stats["L"], self.app.get_color_depth()) # Number of histogram bins.
    self.plotcontrast = False # Plot contrast (instead of stretch) function.
    self.stretchpoints = min(1024, self.histbins) # Number of points on the stretch/contrast function plot.
    self.plot_reference_histograms()
    self.plot_image_histograms()
    self.outofrange = self.reference.is_out_of_range() # Is the reference image out-of-range ?
    if self.outofrange: print("Reference image is out-of-range...")
    self.currentparams = self.get_params()
    self.app.mainwindow.set_rgb_luma_callback(self.update_rgb_luma)
    self.start(identity = not self.outofrange) # If so, the stretch tool may clip the image whatever the parameters.
    return True

  def options_widgets(self, widgets):
    """Return a box with tool options widgets and store the reference to these widgets in container 'widgets'.
       Return None if there are no tool options widgets.
       Must be defined (if needed) in each subclass."""
    return None

  def tab_widgets(self, key, widgets):
    """Return a box with tab widgets for channel 'key' in "R" (red), "G" (green), "B" (blue), "S" (saturation), "V" (value) or "L" (luma),
       and store the reference to these widgets in container 'widgets'.
       Return None if there is no tab for this channel.
       Must be defined (if needed) in each subclass."""
    return None

  # Tool methods.

  def get_params(self):
    """Return tool parameters.
       Must be defined (if needed) in each subclass."""
    return None

  def set_params(self, params):
    """Set tool parameters 'params'.
       Must be defined (if needed) in each subclass."""
    return

  def run(self, params):
    """Run tool for parameters 'params'.
       Must be defined (if needed) in each subclass."""
    return None, False

  def operation(self, params):
    """Return tool operation string for parameters 'params'.
       Must be defined (if needed) in each subclass."""
    return None

  def update_gui(self):
    """Update main window and image histogram."""
    if not self.opened: return
    self.image.stats = self.image.statistics(channels = self.statchannels)
    self.update_image_histograms()
    self.widgets.fig.canvas.draw_idle()
    super().update_gui()

  # Plot histograms, stretch function, display stats...

  def plot_reference_histograms(self):
    """Plot reference histograms."""
    edges, counts = self.reference.histograms(channels = self.histchannels, nbins = self.histbins)
    ax = self.widgets.fig.add_subplot(211)
    self.widgets.fig.refhistax = ax
    ax.histlines = plot_histograms(ax, edges, counts, colors = self.histcolors,
                                   title = "Reference", xlabel = None, ylogscale = self.histlogscale)
    tmin = min(0., edges[0]) # Initialize stretch function plot.
    tmax = max(1., edges[1])
    t = np.linspace(tmin, tmax, int(round(self.stretchpoints*(tmax-tmin))))
    self.widgets.fig.stretchax = self.widgets.fig.refhistax.twinx()
    ax = self.widgets.fig.stretchax
    ax.clear()
    ax.stretchline, = ax.plot(t, t, linestyle = ":", zorder = -1)
    ax.diagline,    = ax.plot([0., 1.], [0., 1.], color = "gray", linestyle = ":", linewidth = 1., zorder = -3)
    ax.zeroline     = ax.axhline(0., color = "gray", linestyle = ":", linewidth = 1., zorder = -3)
    ax.yaxis.set_label_position("right")
    self.update_stretch_function_axes()
    self.add_histogram_widgets(ax)

  def update_reference_histograms(self):
    """Update reference histograms."""
    edges, counts = self.reference.histograms(channels = self.histchannels, nbins = self.histbins)
    ax = self.widgets.fig.refhistax
    update_histograms(ax, ax.histlines, edges, counts, ylogscale = self.histlogscale)

  def plot_image_histograms(self):
    """Plot image histograms."""
    ax = self.widgets.fig.add_subplot(212)
    self.widgets.fig.imghistax = ax
    edges, counts = self.image.histograms(channels = self.histchannels, nbins = self.histbins)
    ax.histlines = plot_histograms(ax, edges, counts, colors = self.histcolors,
                                   title = "Image", ylogscale = self.histlogscale)

  def update_image_histograms(self):
    """Update image histograms."""
    edges, counts = self.image.histograms(channels = self.histchannels, nbins = self.histbins)
    ax = self.widgets.fig.imghistax
    update_histograms(ax, ax.histlines, edges, counts, ylogscale = self.histlogscale)
    tab = self.widgets.rgbtabs.get_current_page()
    key = self.channelkeys[tab]
    self.display_stats(key)

  def update_stretch_function_axes(self):
    """Update stretch function plot axis (switch between stretch function and contrast enhancement plots)."""
    ax = self.widgets.fig.stretchax
    if self.plotcontrast:
      ax.diagline.set_visible(False)
      ax.zeroline.set_visible(True)
      ax.set_ylabel("CE function log(f')")
    else:
      ax.diagline.set_visible(True)
      ax.zeroline.set_visible(False)
      ax.set_ylabel("Stretch function f")
      ax.set_ylim(0., 1.)

  def plot_stretch_function(self, f, color):
    """Plot the stretch function f or the contrast enhancement function log(f') with color 'color'."""
    ax = self.widgets.fig.stretchax
    line = ax.stretchline
    t = line.get_xdata()
    ft = f(t)
    if self.plotcontrast: # Contrast enhancement function.
      ft = np.log(np.maximum(np.gradient(ft, t), 1.e-12))
      ymin = ft[ft > np.log(1.e-12)].min()
      ymax = ft.max()
      dy = ymax-ymin
      ax.set_ylim(ymin-.025*dy, ymax+.025*dy)
    line.set_ydata(ft)
    line.set_color(color)

  def display_stats(self, key):
    """Display reference and image statistics for channel 'key'."""
    self.widgets.refstats.set_label(stats_string(self.reference.stats[key]))
    self.widgets.imgstats.set_label(stats_string(self.image.stats[key]))

  def stretch_function(self, t, params):
    """Return the stretch function f(t) for parameters 'params'.
       Must be defined (if needed) in each subclass."""
    return t

  def add_histogram_widgets(self, ax):
    """Add histogram widgets (other than stretch function) in axes 'ax'.
       Must be defined (if needed) in each subclass."""
    return

  # Update histograms, stats... on widget or key_press events.

  def update(self, changed, **kwargs):
    """Update histograms, stats and widgets."""
    if changed == "tab":
      tab = kwargs["tab"]
      key = self.channelkeys[tab]
      idx = self.histchannels.index(key)
      highlight_histogram(self.widgets.fig.refhistax.histlines, idx)
      highlight_histogram(self.widgets.fig.imghistax.histlines, idx)
      self.display_stats(key)
    else:
      tab = self.widgets.rgbtabs.get_current_page()
      key = self.channelkeys[tab]
    if changed is not None:
      self.update_widgets(key, changed)
      self.widgets.fig.canvas.draw_idle()
    self.reset_polling(self.get_params()) # Expedite main window update.

  def update_widgets(self, key, changed):
    """Update widgets (other than histograms and stats) on change of 'changed' in channel 'key'.
       Must be defined (if needed) in each subclass."""
    return

  def key_press(self, widget, event):
    """Callback for key press in the stretch tool window."""
    key = decode_key(event)
    if not key.ctrl and not key.alt:
      if key.uname == "L": # Toggle log scale.
        self.histlogscale = not self.histlogscale
        self.update_reference_histograms()
        self.update_image_histograms()
        self.widgets.fig.canvas.draw_idle()
        self.window.queue_draw()
        return
      elif key.uname == "C": # Toggle stretch function/contrast enhancement function.
        self.plotcontrast = not self.plotcontrast
        tab = self.widgets.rgbtabs.get_current_page()
        key = self.channelkeys[tab]
        self.update_stretch_function_axes()
        self.update_widgets(key, "sfplot")
        self.widgets.fig.canvas.draw_idle()
        self.window.queue_draw()
        return
    super().key_press(widget, event)

  # Callbacks on luma RGB components update in main window.

  def update_rgb_luma(self, rgbluma):
    """Update luma rgb components."""
    self.reference.stats = self.image.statistics(channels = self.statchannels)
    self.update_reference_histograms()
    self.image.stats = self.image.statistics(channels = self.statchannels)
    self.update_image_histograms()
    self.widgets.fig.canvas.draw_idle()
    self.window.queue_draw()



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated (+).

"""Threshold mask tool."""

from ..gtk.customwidgets import VBox, RadioButtons, HScaleSpinButton
from ..toolmanager import BaseToolWindow
from ...imageprocessing import imageprocessing
from skimage.morphology import isotropic_dilation, disk
from scipy.ndimage import convolve, median_filter, maximum_filter, gaussian_filter
import numpy as np

class ThresholdMaskTool(BaseToolWindow):
  """Threshold Mask tool class."""

  _action_ = "Setting-up threshold mask..."

  _help_ = """Set-up a mask where a given convolution filter is greater than a threshold. This filter can be applied to the value, luma or lightness channel, and is characterized by a radius r:
  \u2022 Mean: average of the channel in a disk of radius r.
  \u2022 Median: median of the channel in a disk of radius r.
  \u2022 Gaussian: gaussian average of the channel with standard deviation r/3.
  \u2022 Maximum: maximum of the channel in a disk of radius r.
Once computed, the mask can be extended and smoothed over the given lengths."""

  _onthefly_ = False # This transformation can not be applied on the fly.

  DARKCOLOR  = np.array([[.5], [0.], [.5]], dtype = imageprocessing.IMGTYPE)
  LIGHTCOLOR = np.array([[0.], [.5], [0.]], dtype = imageprocessing.IMGTYPE)

  def open(self, image):
    """Open tool window for image 'image'."""
    if not super().open(image, "Threshold mask"): return False
    wbox = VBox()
    self.window.add(wbox)
    self.widgets.channelbuttons = RadioButtons(("V", "HSV value"), ("L", "Luma"), ("L*", "Lightness L*"))
    wbox.pack(self.widgets.channelbuttons.hbox(prepend = "Filter channel:"))
    self.widgets.functionbuttons = RadioButtons(("gaussian", "Gaussian"), ("mean", "Mean"), ("median", "Median"), ("maximum", "Maximum"))
    wbox.pack(self.widgets.functionbuttons.hbox(prepend = "Filter function:"))
    self.widgets.radiusscale = HScaleSpinButton(8., 1., 50., 1., digits = 0, length = 480)
    wbox.pack(self.widgets.radiusscale.layout2("Filter radius (pixels):"))
    self.widgets.thresholdscale = HScaleSpinButton(0., 0., 1., .001, digits = 3, length = 480)
    wbox.pack(self.widgets.thresholdscale.layout2("Threshold:"))
    self.widgets.extendscale = HScaleSpinButton(0., 0., 100., 1., digits = 0, length = 480)
    wbox.pack(self.widgets.extendscale.layout2("Extend mask by (pixels):"))
    self.widgets.smoothscale = HScaleSpinButton(0., 0., 100., 1., digits = 0, length = 480)
    wbox.pack(self.widgets.smoothscale.layout2("Smooth mask over (pixels):"))
    wbox.pack(self.tool_control_buttons())
    self.fparams = None
    self.start(identity = False)
    return True

  def get_params(self):
    """Return tool parameters."""
    fchannel = self.widgets.channelbuttons.get_selected()
    ffunction = self.widgets.functionbuttons.get_selected()
    fradius = int(round(self.widgets.radiusscale.get_value()))
    threshold = self.widgets.thresholdscale.get_value()
    extend = int(round(self.widgets.extendscale.get_value()))
    smooth = int(round(self.widgets.smoothscale.get_value()))
    rgbluma = imageprocessing.get_rgb_luma() if fchannel == "L" else None
    return fchannel, ffunction, fradius, threshold, extend, smooth, rgbluma

  def set_params(self, params):
    """Set tool parameters 'params'."""
    fchannel, ffunction, fradius, threshold, extend, smooth, rgbluma = params
    self.widgets.channelbuttons.set_selected(fchannel)
    self.widgets.functionbuttons.set_selected(ffunction)
    self.widgets.radiusscale.set_value(fradius)
    self.widgets.thresholdscale.set_value(threshold)
    self.widgets.extendscale.set_value(extend)
    self.widgets.smoothscale.set_value(smooth)

  def run(self, params):
    """Run tool for parameters 'params'."""
    fchannel, ffunction, fradius, threshold, extend, smooth, rgbluma = params
    fparams = (fchannel, ffunction, fradius, rgbluma)
    # Compute the filter if needed.
    if fparams != self.fparams:
      if fchannel == "V":
        channel = self.reference.value()
      elif fchannel == "L":
        channel = self.reference.luma()
      else:
        channel = self.reference.srgb_lightness()/100.
      if ffunction == "gaussian":
        self.filtered = gaussian_filter(channel, sigma = fradius/3., mode = "reflect")
      elif ffunction == "mean":
        kernel = disk(fradius, dtype = imageprocessing.IMGTYPE)
        kernel /= np.sum(kernel)
        self.filtered = convolve(channel, kernel, mode = "reflect")
      elif ffunction == "median":
        self.filtered = median_filter(channel, footprint = disk(fradius), mode = "reflect")
      else:
        self.filtered = maximum_filter(channel, footprint = disk(fradius), mode = "reflect")
      self.fparams = fparams
    # Threshold the filter.
    lightmask = (self.filtered >= threshold)
    # Extend the light mask.
    if extend > 0: lightmask = isotropic_dilation(lightmask, extend)
    # Display light and dark masks.
    lightmasked = self.reference.clone()
    lightmasked.rgb[:, lightmask] = self.LIGHTCOLOR
    darkmasked = self.reference.clone()
    darkmasked.rgb[:, ~lightmask] = self.DARKCOLOR
    self.queue_gui_mainloop(self.update_mask_tabs, lightmasked, darkmasked) # Thread-safe.
    # Smooth the light mask.
    mask = lightmask.astype(imageprocessing.IMGTYPE)
    if smooth > 0:
      kernel = disk(smooth, dtype = imageprocessing.IMGTYPE)
      kernel /= np.sum(kernel)
      mask = convolve(mask, kernel, mode = "reflect")
    self.image.rgb[:] = mask
    return params, True

  def operation(self, params):
    """Return tool operation string for parameters 'params'."""
    fchannel, ffunction, fradius, threshold, extend, smooth, rgbluma = params
    if fchannel == "L": fchannel = f"L({rgbluma[0]:.2f}, {rgbluma[1]:.2f}, {rgbluma[2]:.2f})"
    return f"ThresholdMask(channel = {fchannel}, filter = {ffunction}, radius = {fradius} pixels, threshold = {threshold:.3f}, extend = {extend} pixels, smooth = {smooth} pixels)"

  def cancel(self, *args, **kwargs):
    """Cancel tool."""
    self.app.mainwindow.delete_image("Light mask", force = True, failsafe = True)
    self.app.mainwindow.delete_image("Dark mask", force = True, failsafe = True)
    super().cancel()

  def cleanup(self):
    """Free memory on exit."""
    try:
      del self.filtered
    except:
      pass

  # Update mask tabs.

  def update_mask_tabs(self, lightmasked, darkmasked):
    """Update light mask tab with image 'lightmasked' and dark mask tab with image 'darkmasked'."""
    self.app.mainwindow.update_image("Light mask", lightmasked, create = True)
    self.app.mainwindow.update_image("Dark mask", darkmasked, create = True)



# This program is 0free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17

"""Image processing definitions."""

import numpy as np
from PIL import Image as PILImage

# Use imageio or imread module ?

IMAGEIO = False

# Image formats (unless otherwise specified):
# -------------------------------------------
#
#  - RGB images are stored as a np.ndarray with dimensions (3, width, height).
#    The first component is red, the second green, and the third blue.
#    They are floats within [0, 1] and type IMGTYPE = np.float32 or np.float64.
#
#  - HSV images are stored as a np.ndarray with dimensions (width, height, 3).
#    The first component is the hue, the second the saturation, and the third the value.
#    They are floats within [0, 1] and type IMGTYPE = np.float32 or np.float64.

# Data type used for images (either np.float32 or np.float64).

IMGTYPE = np.float32

# Expected accuracy in np.float32/np.float64 calculations.

IMGTOL = 1.e-6 if IMGTYPE is np.float32 else 1.e-9

# Image resampling methods, imported from PIL.

NEAREST  = PILImage.Resampling.NEAREST
BILINEAR = PILImage.Resampling.BILINEAR
BICUBIC  = PILImage.Resampling.BICUBIC
LANCZOS  = PILImage.Resampling.LANCZOS
BOX      = PILImage.Resampling.BOX
HAMMING  = PILImage.Resampling.HAMMING




# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated (+).

"""Remove hot pixels tool."""

from ..gtk.customwidgets import VBox, RadioButtons, SpinButton
from ..toolmanager import BaseToolWindow
from ...imageprocessing import imageprocessing

class RemoveHotPixelsTool(BaseToolWindow):
  """Remove hot pixels tool window class."""

  _action_ = "Removing hot pixels..."

  _help_ = """Replace "hot" pixels > ratio*(average of the 8 neighbors) by this average.
The operation can be applied to the red, blue and green channels independently, or to the luma."""

  def open(self, image):
    """Open tool window for image 'image'."""
    if not super().open(image, "Remove hot pixels"): return False
    wbox = VBox()
    self.window.add(wbox)
    self.widgets.channelbuttons = RadioButtons(("RGB", "RGB"), ("L", "Luma"))
    wbox.pack(self.widgets.channelbuttons.hbox(prepend = "Channel(s):"))
    self.widgets.ratiospin = SpinButton(2., 1., 10., .01, digits = 2)
    wbox.pack(self.widgets.ratiospin.hbox(prepend = "Ratio:"))
    wbox.pack(self.tool_control_buttons(reset = not self.onthefly))
    if self.onthefly:
      self.connect_update_request(self.widgets.channelbuttons, "toggled")
      self.connect_update_request(self.widgets.ratiospin, "value-changed")
    self.start(identity = False)
    return True

  def get_params(self):
    """Return tool parameters."""
    return self.widgets.channelbuttons.get_selected(), self.widgets.ratiospin.get_value(), imageprocessing.get_rgb_luma()

  def set_params(self, params):
    """Set tool parameters 'params'."""
    channels, ratio, rgbluma = params
    self.widgets.channelbuttons.set_selected(channels)
    self.widgets.ratiospin.set_value(ratio)

  def run(self, params):
    """Run tool for parameters 'params'."""
    channels, ratio, rgbluma = params
    self.image.copy_image_from(self.reference)
    self.image.remove_hot_pixels(ratio, channels = channels)
    return params, True

  def operation(self, params):
    """Return tool operation string for parameters 'params'."""
    channels, ratio, rgbluma = params
    if channels == "L": channels = f"L({rgbluma[0]:.2f}, {rgbluma[1]:.2f}, {rgbluma[2]:.2f})"
    return f"RemoveHotPixels(channels = {channels}, ratio = {ratio:.2f})"



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated (+).

"""Color balance tool."""

from ..gtk.customwidgets import HBox, VBox, SpinButton
from ..toolmanager import BaseToolWindow

class ColorBalanceTool(BaseToolWindow):
  """Color balance tool class."""

  _action_ = "Balancing colors..."

  _help_ = "Scale the red, blue and green channels by the given factors."

  def open(self, image):
    """Open tool window for image 'image'."""
    if not super().open(image, "Color balance"): return False
    wbox = VBox()
    self.window.add(wbox)
    hbox = HBox()
    wbox.pack(hbox)
    self.widgets.redspin = SpinButton(1., 0., 2., .01, digits = 2)
    hbox.pack("Red:")
    hbox.pack(self.widgets.redspin)
    self.widgets.greenspin = SpinButton(1., 0., 2., .01, digits = 2)
    hbox.pack(8*" "+"Green:")
    hbox.pack(self.widgets.greenspin)
    self.widgets.bluespin = SpinButton(1., 0., 2., .01, digits = 2)
    hbox.pack(8*" "+"Blue:")
    hbox.pack(self.widgets.bluespin)
    wbox.pack(self.tool_control_buttons())
    if self.onthefly:
      self.connect_update_request(self.widgets.redspin  , "value-changed")
      self.connect_update_request(self.widgets.greenspin, "value-changed")
      self.connect_update_request(self.widgets.bluespin , "value-changed")
    self.start(identity = True)
    return True

  def get_params(self):
    """Return tool parameters."""
    return self.widgets.redspin.get_value(), self.widgets.greenspin.get_value(), self.widgets.bluespin.get_value()

  def set_params(self, params):
    """Set tool parameters 'params'."""
    red, green, blue = params
    self.widgets.redspin.set_value(red)
    self.widgets.greenspin.set_value(green)
    self.widgets.bluespin.set_value(blue)

  def run(self, params):
    """Run tool for parameters 'params'."""
    red, green, blue = params
    self.image.copy_image_from(self.reference)
    transformed = (red != 1. or green != 1. or blue != 1.)
    if transformed: self.image.color_balance(red, green, blue)
    return params, transformed

  def operation(self, params):
    """Return tool operation string for parameters 'params'."""
    red, green, blue = params
    return f"ColorBalance(R = {red:.2f}, G = {green:.2f}, B = {blue:.2f})"



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17

"""Pixel math framework."""

import re
from .defs import IMGTYPE
from . import utils
from . import colors
from . import stretchfunctions as stretch

class PixelMath:
  """Pixel math class."""

  def __init__(self, images):
    """Initialize the class with the set of images 'images'."""
    self.images = images

  def run(self, command):
    """Run pixel math command 'command' and return the result."""

    globs = {"__builtins__": {}} # Hide all globals including python builtins for (minimal) security.

    # Set-up command environment.

    import numpy as np

    def value(image, midtone = .5):
      """Return the HSV value of image 'image' with midtone correction 'midtone'."""
      return mts(colors.hsv_value(image), midtone)

    def luma(image, midtone = .5):
      """Return the luma of image 'image' with midtone correction 'midtone'."""
      return mts(colors.luma(image), midtone)

    def luminance(image, midtone = .5):
      """Return the luminance of image 'image' with midtone correction 'midtone'."""
      return mts(colors.lrgb_to_srgb(colors.srgb_luminance(image)), midtone)

    def lightness(image, midtone = .5):
      """Return the CIE lightness of image 'image' with midtone correction 'midtone'."""
      return mts(colors.srgb_lightness(image), midtone)

    def scale(image, source, target):
      """Scale image 'image' by the ratio 'target'/'source'."""
      return utils.scale_pixels(image, source, target)

    def blend(image1, image2, mix):
      """Blend images 'image1' and 'image2' as (1-mix)*image1+mix*image2."""
      return (1.-mix)*image1+mix*image2

    def mts(image, midtone):
      """Apply midtone stretch function with midtone 'midtone' to image 'image'."""
      return (midtone-1.)*image/((2.*midtone-1.)*image-midtone) if midtone != .5 else image

    def ghs(image, lnD1, B, SYP, SPP = 0., HPP = 1.):
      """Apply generalized hyperbolic stretch function with parameters (lnD1, B, SYP, SPP,HPP) to image 'image'."""
      return stretch.ghyperbolic_stretch_function(image, (lnD1, B, SYP, SPP, HPP, False))

    # Register the environment as globals.

    globs.update({"np": np, "value": value, "luma": luma, "luminance": luminance, "lightness": lightness, "scale": scale, "blend": blend, "mts": mts, "ghs": ghs})

    # Register all images as locals.

    locls = {f"IMG{n+1}": self.images[n].rgb for n in range(len(self.images))}

    # Execute the command and return the result converted to IMGTYPE.

    return IMGTYPE(eval(command, globs, locls))



# This program is 0free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17

"""Color space management."""

import numpy as np
import matplotlib.colors as mplcolors
from .defs import IMGTYPE, IMGTOL

#########
# Luma. #
#########

# Weight of the RGB components in the luma.

rgbluma = IMGTYPE((.3, 0.6, 0.1))

def get_rgb_luma():
  """Return the weights of the RGB components in the luma."""
  return tuple(rgbluma)

def set_rgb_luma(rgb):
  """Set the weights 'rgb' of the RGB components in the luma."""
  global rgbluma
  rgbluma = IMGTYPE(rgb)

def luma(image):
  """Return the luma of the RGB image 'image',
     defined as the linear combination of the RGB components weighted by rgbluma."""
  return rgbluma[0]*image[0]+rgbluma[1]*image[1]+rgbluma[2]*image[2]

#############################
# sRGB <-> lRGB conversion. #
#############################

def srgb_to_lrgb(image):
  """Convert the sRGB image 'image' into a linear RGB image."""
  srgb = np.clip(image, 0., 1.)
  return np.where(srgb > .04045, ((srgb+0.055)/1.055)**2.4, srgb/12.92)

def lrgb_to_srgb(image):
  """Convert the linear RGB image 'image' into a sRGB image."""
  lrgb = np.clip(image, 0., 1.)
  return np.where(lrgb > .0031308, 1.055*lrgb**(1./2.4)-0.055, 12.92*lrgb)

############################
# Luminance and lightness. #
############################

def lrgb_luminance(image):
  """Return the luminance Y of the linear RGB image 'image'."""
  return .2126*image[0]+0.7152*image[1]+0.0722*image[2]

def lrgb_lightness(image):
  """Return the CIE lightness L* of the linear RGB image 'image'.
     Warning: L* is defined within [0, 100] instead of [0, 1]."""
  Y = lrgb_luminance(image)
  return np.where(Y > .008856, 116.*Y**(1./3.)-16., 903.3*Y)

def srgb_luminance(image):
  """Return the luminance Y of the sRGB image 'image'."""
  return lrgb_luminance(srgb_to_lrgb(image))

def srgb_lightness(image):
  """Return the CIE lightness L* of the sRGB image 'image'.
     Warning: L* is defined within [0, 100] instead of [0, 1]."""
  return lrgb_lightness(srgb_to_lrgb(image))

###########################
# RGB <-> HSV conversion. #
###########################

def hsv_value(image):
  """Return the HSV value = max(RGB) of the RGB image 'image'."""
  return image.max(axis = 0)

def hsv_saturation(image):
  """Return the HSV saturation = 1-min(RGB)/max(RGB) of the RGB image 'image'."""
  return 1.-image.min(axis = 0)/image.max(axis = 0, initial = IMGTOL) # Safe evaluation.

def rgb_to_hsv(image):
  """Convert the RGB image 'image' into a HSV image.
     Warning: The HSV components are returned as an array with dimensions (height, width, 3)."""
  return IMGTYPE(mplcolors.rgb_to_hsv(np.moveaxis(image, 0, -1)))

def hsv_to_rgb(image):
  """Convert the HSV image 'image' into a RGB image.
     Warning: The HSV components 'image' are input as an array with dimensions (height, width, 3)."""
  return IMGTYPE(np.moveaxis(mplcolors.hsv_to_rgb(image), -1, 0))



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated (+).

"""Gray scale conversion tool."""

from ..gtk.customwidgets import VBox, RadioButtons
from ..toolmanager import BaseToolWindow
from ...imageprocessing import imageprocessing

class GrayScaleConversionTool(BaseToolWindow):
  """Gray scale conversion tool window class."""

  _action_ = "Converting into a gray scale image..."

  _help_ = """Convert into a gray scale image with same HSV value, luma or luminance/lightness."""

  def open(self, image):
    """Open tool window for image 'image'."""
    if not super().open(image, "Gray scale conversion"): return False
    wbox = VBox()
    self.window.add(wbox)
    self.widgets.channelbuttons = RadioButtons(("V", "HSV Value"), ("L", "Luma"), ("Y", "Luminance Y"))
    wbox.pack(self.widgets.channelbuttons.hbox(prepend = "Channel:"))
    wbox.pack(self.tool_control_buttons(reset = False))
    if self.onthefly:
      self.connect_update_request(self.widgets.channelbuttons, "toggled")
    self.start(identity = False)
    return True

  def get_params(self):
    """Return tool parameters."""
    return self.widgets.channelbuttons.get_selected(), imageprocessing.get_rgb_luma()

  def set_params(self, params):
    """Set tool parameters 'params'."""
    channel, rgbluma = params
    self.widgets.channelbuttons.set_selected(channel)

  def run(self, params):
    """Run tool for parameters 'params'."""
    channel, rgbluma = params
    self.image.copy_image_from(self.reference)
    self.image.gray_scale(channel = channel)
    #lightdiff = self.image.srgb_lightness()-self.reference.srgb_lightness()
    #print(f"Maximum lightness difference = {abs(lightdiff).max()}.")
    return params, True

  def operation(self, params):
    """Return tool operation string for parameters 'params'."""
    channel, rgbluma = params
    if channel == "L": channel = f"L({rgbluma[0]:.2f}, {rgbluma[1]:.2f}, {rgbluma[2]:.2f})"
    return f"GrayScale(channel = {channel})"



# This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
# Author: Yann-Michel Niquet (contact@ymniquet.fr).
# Version: 1.8.0 / 2025.12.17
# GUI updated (+).

"""Edition with external tools."""

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, GObject
from .gtk.customwidgets import Label, VBox, HButtonBox, Button, RadioButtons, Entry
from .base import BaseWindow, Container, InfoDialog, ErrorDialog
import os
import tempfile
import threading
import subprocess

class EditTool(BaseWindow):
  """External editor class."""

  def __init__(self, app, command = None, filename = "eQuimage.tiff", depth = 32):
    """Initialize external editor for app 'app'.
        - 'command' is the command to be run (e.g., "gimp -n $"). "$" is replaced by the name of the image file to be opened by the editor.
        - 'filename' is the name of the image file to be opened by the editor, created in a temporary directory.
        - 'depth' is the default color depth of this file."""
    super().__init__(app)
    self.set_command(command)
    self.set_filename(filename, depth)

  def set_command(self, command):
    """(Re)Set editor command 'command'."""
    self.command = command

  def set_filename(self, filename, depth = 32):
    """(Re)Set image file name 'filename' and color depth 'depth'."""
    self.filename = filename
    self.set_depth(depth)

  def set_depth(self, depth):
    """(Re)Set image file color depth 'depth'."""
    self.depth = depth

  def open_window(self, title = "Edit with..."):
    """Open a modal Gtk window with title 'title'. This window must remain open while running the editor.
       Return window and widgets container."""
    self.opened = True
    self.window = Gtk.Window(title = title,
                             transient_for = self.app.mainwindow.window,
                             modal = True,
                             border_width = 16)
    self.window.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)
    self.window.connect("delete-event", self.close_window)
    self.widgets = Container()
    self.widgets.commententry = None
    self.widgets.editbutton = None
    return self.window, self.widgets

  def close_window(self, *args, **kwargs):
    """Close tool window."""
    self.window.destroy()
    self.opened = False
    del self.widgets

  def comment_entry(self):
    """Add a comment entry to the tool widgets.
       Return an Entry object, which must be packed appropriately in the tool window."""
    self.widgets.commententry = Entry(text = "", width = 64)
    return self.widgets.commententry

  def edit_cancel_buttons(self):
    """Add edit & cancel buttons to the tool widgets.
       Return a HButtonBox object, which must be packed appropriately in the tool window."""
    hbox = HButtonBox()
    self.widgets.editbutton = Button(label = "Edit")
    self.widgets.editbutton.connect("clicked", lambda button: self.edit())
    hbox.pack(self.widgets.editbutton)
    self.widgets.cancelbutton = Button(label = "Cancel")
    self.widgets.cancelbutton.connect("clicked", lambda button: self.close_window())
    hbox.pack(self.widgets.cancelbutton)
    return hbox

  def edit(self):
    """Run editor on current image."""

    def run():
      """Run editor in a separate thread."""

      def finalize(image, msg = None, error = False):
        """Finalize run (register image 'image', close window and open info/error dialog with message 'msg')."""
        if image is not None:
          if self.widgets.commententry is not None:
            comment = self.widgets.commententry.get_text().strip()
            if len(comment) > 0: comment = " # "+comment
          else:
            comment = ""
          self.app.finalize_tool(image, f"Edit('{editor}'){comment}")
        self.close_window()
        if msg is not None:
          Dialog = ErrorDialog if error else InfoDialog
          Dialog(self.app.mainwindow.window, str(msg))
        return False

      try:
        with tempfile.TemporaryDirectory() as tmpdir:
          # Set tmp file name.
          tmpfile = os.path.join(tmpdir, self.filename)
          # Process command.
          cmd = self.command.strip() if self.command is not None else ""
          if cmd == "": raise RuntimeError("Please specify editor command.")
          splitcmd = []
          filefound = False
          for item in cmd.split(" "):
            if item == "$":
              filefound = True
              splitcmd.append(tmpfile)
            else:
              splitcmd.append(item)
          if not filefound: raise RuntimeError("No place holder for the image file ($) found in the editor command.")
          editor = os.path.basename(splitcmd[0])
          # Save image.
          image = self.app.get_image()
          image.save(tmpfile, depth = self.depth)
          ctime = os.path.getmtime(tmpfile)
          # Run editor.
          print(f"Editing with {editor}...")
          subprocess.run(splitcmd)
          if self.opened: # Cancel operation if the window has been closed in the meantime.
            # Check if the image has been modified by the editor.
            mtime = os.path.getmtime(tmpfile)
            if mtime != ctime: # If so, load and register the new one...
              print(f"The file {tmpfile} has been modified by {editor}; Reloading in eQuimageGtk...")
              image = self.app.ImageClass()
              image.load(tmpfile)
              if not image.is_valid(): raise RuntimeError(f"The image returned by {editor} is invalid.")
              GObject.idle_add(finalize, image, None, False)
            else: # Otherwise, open info dialog and cancel operation.
              print(f"The file {tmpfile} has not been modified by {editor}; Cancelling operation...")
              GObject.idle_add(finalize, None, f"The image has not been modified by {editor}.\nCancelling operation.", False)
      except Exception as err:
        GObject.idle_add(finalize, None, err, True)

    if self.widgets.editbutton is not None:
      self.widgets.editbutton.set_sensitive(False) # Can only be run once.
    thread = threading.Thread(target = run, daemon = False)
    thread.start()

#
# Applications: siril, gimp, etc...
#

def edit_with_siril(app):
  """Edit current image of app 'app' with siril."""
  if not app.get_context("image"): return
  editor = EditTool(app, "siril $", "eQuimage.fits", depth = 32)
  window, widgets = editor.open_window(title = "Edit with siril")
  wbox = VBox()
  window.add(wbox)
  wbox.pack(Label("The image will be saved as a 32 bpc (float) FITS file and edited with siril."))
  wbox.pack(Label("Overwrite the file when leaving siril."))
  wbox.pack(Label("You can enter a comment for the logs below, <b>before</b> closing siril:"))
  wbox.pack(editor.comment_entry().hbox())
  wbox.pack(Label("<b>The operation will be cancelled if you close this window !</b>"))
  wbox.pack(editor.edit_cancel_buttons())
  window.show_all()

def edit_with_gimp(app):
  """Edit current image of app 'app' with gimp."""

  class GimpEditTool(EditTool):
    """Gimp editor child class."""

    def edit(self):
      """Update image color depth and run gimp on current image."""
      depth = self.widgets.depthbuttons.get_selected()
      self.set_depth(depth)
      super().edit()

  if not app.get_context("image"): return
  editor = GimpEditTool(app, "gimp -n $", "eQuimage.tiff")
  window, widgets = editor.open_window(title = "Edit with gimp")
  wbox = VBox()
  window.add(wbox)
  wbox.pack(Label("The image will be saved as a:"))
  widgets.depthbuttons = RadioButtons((8, "8 bpc"), (16, "16 bpc"), (32, "32 bpc"))
  widgets.depthbuttons.set_selected(32)
  wbox.pack(widgets.depthbuttons.hbox())
  wbox.pack(Label("TIFF file and edited with gimp."))
  wbox.pack(Label("Overwrite the file when leaving gimp."))
  wbox.pack(Label("You can enter a comment for the logs below, <b>before</b> closing gimp:"))
  wbox.pack(editor.comment_entry().hbox())
  wbox.pack(Label("<b>The operation will be cancelled if you close this window !</b>"))
  wbox.pack(editor.edit_cancel_buttons())
  window.show_all()

def edit_with_any(app):
  """Edit current image of app 'app' with any editor."""

  class AnyEditTool(EditTool):
    """Any editor child class."""

    def edit(self):
      """Update command, image file name & color depth and run editor on current image."""
      command = self.widgets.commandentry.get_text()
      self.set_command(command)
      suffix, depth = self.widgets.filebuttons.get_selected()
      self.set_filename("eQuimage."+suffix, depth)
      super().edit()

  if not app.get_context("image"): return
  editor = AnyEditTool(app)
  window, widgets = editor.open_window(title = "Edit with...")
  wbox = VBox()
  window.add(wbox)
  wbox.pack(Label("The image will be saved as a:"))
  widgets.filebuttons = RadioButtons((("tiff", 8), "8 bpc TIFF"), (("tiff", 16), "16 bpc TIFF"), (("tiff", 32), "32 bpc TIFF"), (("fits", 32), "32 bpc (float) FITS"))
  widgets.filebuttons.set_selected(("tiff", 32))
  wbox.pack(widgets.filebuttons.hbox())
  wbox.pack(Label("file and edited with (type command below and use \"$\" as a place holder\nfor the image file name):"))
  widgets.commandentry = Entry(text = "gimp -n $", width = 64)
  wbox.pack(widgets.commandentry.hbox())
  wbox.pack(Label("Overwrite the file when leaving the editor."))
  wbox.pack(Label("You can enter a comment for the logs below, <b>before</b> closing the editor:"))
  wbox.pack(editor.comment_entry().hbox())
  wbox.pack(Label("<b>The operation will be cancelled if you close this window !</b>"))
  wbox.pack(editor.edit_cancel_buttons())
  window.show_all()
